;;; -*- Syntax: common-lisp; Mode: LISP; Package: (parser :use (parser-util common-lisp)); Base: 10 -*- 

;;; This file is a first shot at converting English syntax into Montague syntax

(in-package 'parser :use '(parser-util screamer lisp))


;;; a complex sentence is one that need more work than just stripping the 
;;; extra info in order to be in proper Montague Syntax format

;;; the set of constructors that make up "complex sentences"
(setq *complex-sentences-constructors* 
      '(trace np-num tied-to-np-num tied-to-np-number  "*0*" "*1*" "*2*"
	      "*3*" "*4*" "*5*" "*6*" "*7*" "*8*" "*9*"))

(defun spec-member? (symb lst)
    (cond ((not (listp lst))
	   (error "~%Error: ~s must be of type list~%" lst))
	  ((and (atom symb)
		(not (stringp symb))
		(listp lst))
	   (member symb lst))
	  ((stringp symb)
	   (string-member symb lst))))

(defun string-member (symb lst)
    (if (not (null lst))
	(or (string-equal symb (car lst))
	    (string-member symb (cdr lst)))))

(defun parsed-sentence->montague (parsed-sent sent)
  (format t "~%~%Montague syntax:~%")
  (format t "~s~%" 
	  (convert-all
	    (strip-extra-parens
	      (remove-traces (strip-extra-parens
			       (strip-extra-info 
				 parsed-sent 
				 (append sent 
					 *complex-sentences-constructors*)))
			     sent)))))

(setf *quantifiers* '(every some))

;;(member (caar lst) *quantifiers*)

(defun continue-conversion (lst type)
  (let ((quant-binding (gensym "X-")))
    (cond ((eq type 'quantifier) 
	   `(,(car lst)
	      (is (some (some-such-that
			  ,(cadr (car lst)) ,quant-binding
			  (,quant-binding 	
			   ,(adjust-quantifiers 
			      (car lst) quant-binding 
			      (cdr lst))))))))
	  ((eq type 'negation)
	   `(not ,(adjust-negation lst)))
	  ((eq type 'intersect)
	   (adjust-intersect lst))
	  ((eq type 'conjunction)
	   (adjust-conjunction lst))
	  (t lst))))

(defun convert-all (sent)
  (mvlet (((new-sent negated)
	   (adjust-negation sent '())))
	 (let ((final-sent 
		 (mapcar (lambda (sentence)
			   (adjust-quantifiers sentence))
			 (adjust-sentence new-sent))))
	   (if negated
	       (mapcar (lambda (sentc)
			 `(not ,sentc))
		       final-sent)
	       final-sent))))
	     
	 

(defun adjust-quantifiers (sent)
  (if (not (null sent))
      (let ((quant-binding (gensym "X-")))
;;	(cond ((and (listp sent)
	;;	  (listp (car sent))
		;;  (not (null (car sent)))
		 ;; (not (null (cdr sent)))
		 ;; (eq (caar sent) 'every))))
	     (mvlet (((new-sent changed)
		      (adjust-quantified-values (car sent)
					       quant-binding
					       (cadr sent)
					       '())))
		    (if changed
			`(,(car sent)
			  (is (some (some-such-that
				      ,(cadr (car sent)) ,quant-binding
				      (,quant-binding ,new-sent)))))
			sent)))))

					 


(defun generate-conjoined-sentences (lst)
  (let ((conjoined-nps (get-conjoined-nps lst)))
    (let ((sent `(,(let ((adjusted-conj (adjust-conjunction (car lst))))
		     (if (and (listp adjusted-conj)
			      (eq (length adjusted-conj) 1))
			 (car adjusted-conj)
			 adjusted-conj))
		  ,(cadr lst))))
      (append `(,sent)
	      (mapcar (lambda (subj-val)
		     (substitute-conjoined-nps subj-val sent))
		   conjoined-nps)))))

(defun substitute-conjoined-nps (subj-val sent)
  (nconc `(,subj-val)
	 (cdr sent)))


(defun get-conjoined-nps (sent)
  (if (not (null sent))
      (cond ((and (listp sent)
		  (listp (car sent))
		  (listp (cadr sent))
		  (not (null (car sent)))
		  (not (null (cadr sent))))
	     (nconc (get-conjoined-nps (car sent))
		    (get-conjoined-nps (cadr sent))))
	    ((and (atom (car sent))
		  (listp (cadr sent))
		  (not (null (cadr sent)))
		  (not (member (car sent) *conjunctions*)))
	     (get-conjoined-nps (cadr sent)))
	    ((and (listp sent)
		  (not (null (cdr sent)))
		  (listp (cadr sent))
		  (not (null (cadr sent)))
		  (member (car sent) *conjunctions*))
	     (nconc `(,(car (second sent)))
		    (get-conjoined-nps (second (second sent)))))
	    ((and (listp sent)
		  (not (null (cdr sent)))
		  (member (car sent) *conjunctions*))
	     (cdr sent)))))
		  
		  

(setf *conjunctions* '(and or but))

(defun adjust-conjunction (lst)
  (if (not (null lst))
      (cond ((symbolp lst)
	     lst)
	    ((and (listp lst)
		  (listp (car lst))
		  (listp (cadr lst))
		  (not (null (car lst)))
		  (not (null (cadr lst))))
	     (nconc (adjust-conjunction (car lst))
		    `(,(adjust-conjunction (cadr lst)))))
	    ((and (atom (car lst))
		  (listp (cadr lst))
		  (not (null (cadr lst)))
		  (not (member (car lst) *conjunctions*)))
	     (nconc `(,(car lst))
		    (adjust-conjunction (cadr lst))))
	    ((and (listp lst)
		  (not (null (cdr lst)))
		  (member (car lst) *conjunctions*))
	     '())
	    (t 
	      (if (not (member lst *conjunctions*))
		  lst)))))


(defun adjust-sentence (sent)
  (mapcar (lambda (sent-val)
	    `(,(car sent-val)
	      ,(adjust-vps (cadr sent-val))))
	  (generate-conjoined-sentences sent)))




(setf *aux* '(will do does have has did would can could))

(defun adjust-vps (vp)
  (if (not (null vp))
      (cond ((and (listp (cdr vp))
		  (not (null (cdr vp)))
		  (listp (cadr vp))
		  (not (null (cadr vp)))
		  (eq (second (cadr vp)) 'than))
	     `(,(combine-symbol-list `(,(car vp) 
				       ,(first (cadr vp)) than)) 
	       ,(third (cadr vp))))
	    ((and (listp (cdr vp))
		  (not (null (cdr vp)))
		  (listp (car vp))
		  (not (null (car vp)))
		  (member (caar vp) *aux*)
		  (eq (length (car vp)) 2))
	     `(,(combine-symbol-list (car vp))
	       ,(cadr vp)))
	    ((and (listp (cdr vp))
		  (not (null (cdr vp)))
		  (listp (car vp))
		  (not (null (car vp)))
		  (get (caar vp) 'verb)
		  (eq (length (car vp)) 2)
		  (eq (caadr (car vp)) 'to))
	     `(,(combine-symbol-list `(,(caar vp) ,@(cadr (car vp))))
	       ,(cadr vp)))
	    (t
	      vp))))


(defun adjust-comp-adj (lst)
  (if (not (null lst))
      (cond ((and (listp (cdr lst))
		  (not (null (cdr lst)))
		  (eq (second (cadr lst)) 'than))
	     `(,(combine-symbol-list `(,(car lst) 
				       ,(first (cadr lst)) than)) 
	       ,(third (cadr lst)))))))

(defun adjust-aux-sent (sent)
  (if (not (null sent))
      (cond ((and (listp (cdr sent))
		  (not (null (cdr sent)))
		  (listp (car sent))
		  (not (null (car sent)))
		  (member (caar sent) *aux+not*)
		  (eq (length (car sent)) 2))
	     `(,(combine-symbol-list (car sent))
	       ,(cadr sent))))))

(defun adjust-verb-to-inf-verb (sent)
  (if (not (null sent))
      (cond ((and (listp (cdr sent))
		  (not (null (cdr sent)))
		  (listp (car sent))
		  (not (null (car sent)))
		  (get (caar sent) 'verb)
		  (eq (length (car sent)) 2)
		  (eq (caadr (car sent)) 'to))
	     `(,(combine-symbol-list `(,(caar sent) ,@(cadr (car sent))))
	       ,(cadr sent))))))


(defun adjust-intersect (lst)
  )


(setf *aux+not* '( not))

(defun adjust-negation (lst negated)
  (if (not (null lst))
      (cond 
	;;((and (listp lst)
	;;	  (null (cdr lst))
	;;	  (listp (car lst)))
	;;     (adjust-negation (car lst)))  
	    ((and (listp lst)
		  (atom (car lst))
		  (listp (cadr lst))
		  (not (null (cadr lst))))
	     (mvlet (((new-sent negated)
		      (adjust-negation (cadr lst) negated)))
		    (returned-values 
		      (nconc (if (not (eq (car lst) 'not))
				 `(,(car lst)))
			     `(,new-sent)) negated)))
	    ((and (listp lst)
		  (eq (length lst) 2)
		  (eq (car lst) 'not))
	     (returned-values (cadr lst) t))   
	    ((and (listp lst)
		  (listp (car lst)))
	     (mvlet (((new-sent1 negated1) (adjust-negation (car lst) 
							    negated)))
		    (mvlet (((new-sent2 negated2) (adjust-negation (cadr lst) 
								   negated1)))
			   (returned-values 
			     (list new-sent1 new-sent2) negated2))))
	    (t
	      (if (not (eq lst 'not))
	  	  (returned-values lst negated))))))



(defun adjust-quantified-values (quant-val quant-binding lst changed)      
  (if (not (null lst))
      (cond ((and (listp lst)
		  (listp (car lst))
		  (null (cdr lst)))
	     (adjust-quantified-values quant-val quant-binding (car lst)
				   changed))
	    ((and (listp lst)
		  ;;(atom (car lst))
		  (listp (second lst)))
	     (mvlet (((new-sent changed)
		      (adjust-quantified-values quant-val
					  quant-binding (cadr lst)
					  changed)))
		    (returned-values
		      (list (car lst) new-sent)
		      changed)))
;;	    ((equal (car lst) quant-val)
;;	     (mvlet (((new-sent changed)
;;		      (adjust-quantified-values quant-val 
;;					  quant-binding (cadr lst) 
;;					  changed)))
;;		    (returned-values 
;;		      (nconc `(,quant-binding) new-sent) t)))
	    ((equal lst quant-val)
	     (returned-values quant-binding t))
	    (t 
	      (returned-values lst changed)))))
	     
	     
  
(defun create-assoc-hash-table (lst)
  (if (and (not (null lst))
	   (listp lst))
      (cond ((and (listp lst)
		  (null (cdr lst)))
	     (create-assoc-hash-table (car lst)))
	    ((and (eq (length lst) 2)
		  (listp (car lst))
		  (member (caar lst) *complex-sentences-constructors*))
	     (progn 
	       (put (caar lst) (str->num (second (car lst)))
		    (second lst))
	       (create-assoc-hash-table (cdr lst))))
	    ((listp lst)
	     (progn 
	       (create-assoc-hash-table (car lst))
	       (create-assoc-hash-table (cdr lst)))))))

(defun remove-traces (lst sent)
  (strip-extra-info 
    (substitute-bound-vars 
      (progn (create-assoc-hash-table lst)
	     lst))
    sent))
	    
	  
(defun substitute-bound-vars (lst)
  (if (not (null lst))
      (cond ((and (listp lst)
		  (null (cdr lst))
		  (not (stringp (car lst)))
		  (listp (car lst)))
	     (substitute-bound-vars (car lst)))
	    ((and (listp lst)
		  (eq (length lst) 2)
		  (not (stringp (second lst)))
		  (listp (second lst))
		  (atom (car lst))
		  (eq (car (second  lst)) 'tied-to-np-num))
	     (let ((np-val (get 'np-num (str->num (second (second lst)))))) 
	       (if (eq (car np-val)
		       'trace)
		   (substitute-bound-vars np-val)
		   np-val)))
	    ((or (and (listp lst)
		      (eq (length lst) 2)
		      (not (stringp (second lst)))
		      (listp (second lst))
		      ;;(atom (car lst))
		      (eq (car (second  lst)) 'tied-to-np-number))
		 (stringp lst))
	     '())
	    ((listp lst)
	     (list
	       (substitute-bound-vars (car lst))
	       (substitute-bound-vars (cdr lst))))
	    (t lst))))

		
 

;;;  This function strips any extra information that might be produced by the 
;;;  parser- it only leaves the original sentence along with its parsed parens
(defun strip-extra-info (lst orig-lst)
  (if (not (null lst))
      (if (listp (car lst))
	  (let ((stripped-lst-car (strip-extra-info (car lst) orig-lst)))
	    (if (not (null stripped-lst-car))
		(cons stripped-lst-car
		      (strip-extra-info (cdr lst) orig-lst))
		(strip-extra-info (cdr lst) orig-lst)))
	  (if (spec-member? (car lst)
		      orig-lst)
	      (cons (car lst) (strip-extra-info (cdr lst) orig-lst))
	      (strip-extra-info (delete (car lst) lst) orig-lst)))
      lst))




;;;  This function strips embeded lists and leaves the appropriate last level
(defun strip-extra-parens (lst)
  (if (and (not (symbolp lst))
	   (null (cdr lst)))
      (strip-extra-parens (car lst))
  (if (embeded-lists lst)
      (list (strip-extra-parens (car lst))
	    (strip-extra-parens (cdr lst)))
      (if (and (equal (special-length lst)
		       1)
	       (listp lst))
	  (car lst)
	  lst))))


(defun special-length (object)
  (if (symbolp object)
      1
      (length object)))


(defun embeded-lists (lst)
  (if (and (not (null lst))
	   (not (symbolp lst)))
      (or (listp (car lst))
	  (embeded-lists (cdr lst)))
      '()))
