;;; -*- Syntax: common-lisp; Mode: LISP; Package: (parser-util :use common-lisp); Base: 10 -*- 

(in-package 'parser-util :use '(parser-util screamer lisp))



;; (defexportmacro mvlet (bindings &body body)
;;   (cond ((null (cdr bindings))
;; 	 `(multiple-value-bind ,(first (first bindings)) 
;; 	      ,(second (first bindings))
;; 	    ,@body))
;; 	(t
;; 	 `(multiple-value-bind ,(first (first bindings)) 
;; 	      ,(second (first bindings))
;; 	    (mvlet ,(rest bindings)
;; 	      ,@body)))))


(defexportmacro mvlet ((bindings) &body body)
  (cond ((null (car bindings))
	 `(progn ,@body))
	(t
	  (let ((name (gensym "INSTANCE-")))
	    `(let* ((,name ,(second bindings))
		    (,(first (first bindings)) (first ,name)))
	       (mvlet ((,(rest (first bindings))
		       (rest ,name)))
		      ,@body))))))

(defexportmacro put (arg1 arg2 arg3)
  `(setf (get ,arg1 ,arg2) ,arg3))


(defexportmacro returned-values (&rest args)
   `(mapcar (lambda (x) x)
	    (list ,@args)))

;; for screamer
(defexportmacro declare-nondet-funcs (lst)
  (if (not (null lst))
      `(progn (forward-declare-nondeterministic ,(car lst))
	     (declare-nondet-funcs ,(cdr lst)))
      t))


(defexportmacro catch-fail (sexp)
    `(one-value (either ,sexp 'fail)))
	
;;;  Special comparison functions that include don't cares
(export '*dont-care*)

(defvar *dont-care* '_)
(setf *dont-care* '_)


(defexportmacro spec-eq? (x y)
  `(if (or (eq ,x *dont-care*)
	   (eq ,y *dont-care*))
       t
       (eq ,x ,y)))




(defexportmacro vector-first (vec)
  `(aref ,vec 0))

(defexportmacro vector-second (vec)
  `(aref ,vec 1))

(defexportmacro vector-third (vec)
  `(aref ,vec 2))

(defexportmacro vector-fourth (vec)
  `(aref ,vec 3))




(defexport match-vectors? (x y)
  (and (spec-eq? (vector-first x)
		 (vector-first y))
       (spec-eq? (vector-second x)
		 (vector-second y))
       (spec-eq? (vector-third x)
		 (vector-third y))
       (cond ((and (eq (length x) 4)
		  (eq (length y) 4))
	      (spec-eq? (vector-fourth x)
			(vector-fourth y)))
	     ((or (eq  (length x) 4)
		  (eq (length y) 4))
	      nil)
	     (t t ))))


;; match? is a special version of equal.. it can compare any element and
;; matches it with a '_ (don't care) as equal.

(defexport match? (x y &optional lst1 lst2)
  (if (and (null lst1) 
	   (null lst2))
      (match? x y x y)
      (cond ;;((not (eq (length lst1)
		;;      (length lst2)))
	    ;; '())
	    ((and (vectorp x)
		  (vectorp y))
	     (match-vectors? x y))
	    ((equal x y)
	     t)
	    ((and (listp x)
		  (listp y))
	     (if (or (eq (car x) *dont-care*)
		     (eq (car y) *dont-care*))
		 (match? (cdr x) (cdr y) lst1 lst2)
		 (and (member (car x) lst2)
		      (member (car y) lst1)
		      (match? (cdr x) (cdr y) lst1 lst2))))
	    ((or (listp x) 
		 (listp y))
	     '())
	    (t
	      (spec-eq? x y)))))


(defexport combine-symbols (s1 s2 &optional (package *package*))
  (intern (concatenate 'string (format nil "~s" s1) "-" (format nil "~s" s2))
	  package))

(defexport combine-symbol-list (s-list &optional (package *package*))
  (cond ((null s-list) nil)
	((null (rest s-list)) (car s-list))
	(t (combine-symbols (car s-list)
			    (combine-symbol-list (cdr s-list) package)
			    package))))