;;; -*- Syntax: common-lisp; Mode: LISP; Package: (parser :use (parser-util common-lisp)); Base: 10 -*- 

;; Sample parse test grammar for an XG

(in-package 'parser :use '(parser-util screamer lisp))

;; to parse a whole paragraph..
(defun parse-par ()
  (format t "~%Enter a paragraph to be parsed (all sentences MUST be between parentheses):~%")
  (let ((paragraph (read)))
    (mapcan (lambda (sentc)
	      (let ((lst (one-value (parse-now sentc) 'fail)))
		(if (not (equal 'fail lst))
		    (progn
		      (format t "~%For the sentence:~%~s~%" sentc)
		      (pprint lst)
		      (parsed-sentence->montague lst sentc))
		    (format t "~%Failed to parse input sentence:~%~s~%" 
			    sentc))))
	    paragraph)))
	  
	    

;; main read-parse-print loop..
(defun parse ()
  (loop
    (format t "~%Enter sentence to be parsed (between parentheses):~%")
    (let ((sentc (read)))
	(format t "lst:~s~%" sentc)
      (let ((lst (one-value (parse-now sentc) 'fail)))
	(format t "lst:~s~%" lst)
	(if (not (equal 'fail lst))
	    (progn
	      (pprint lst)
	      (parsed-sentence->montague lst sentc))
	    (format t "~%Failed to parse input sentence..~%"))))))


(defun-nondeterministic parse-now (input-sentence)
  (let ((result-of-parse (sentence input-sentence top-level-stack)))
    (mvlet (((parsed-sentence rem final-stack)
	     result-of-parse))
	   (if (and (null rem)
		    (null final-stack))
	       parsed-sentence
	       (fail)))))

		   

;; define the an xlist stack to be used throughout the parse
;; define top level empty stack
(defvar top-level-stack '())

;; parse sentence
;; one of four different possiblities
(defun-nondeterministic sentence (sent xlist-stack)
  (sentence1 sent xlist-stack))

;; sentence -> declarative terminator(.)
(defun-nondeterministic sentence1 (sent xlist-stack)
  (mvlet (((parsed-s s-rem final-stack)
	   (mvlet (((parsed-s mask rem1 stack1)
		    (declarative sent xlist-stack)))
		  (returned-values parsed-s rem1 stack1))))
	 (returned-values parsed-s s-rem final-stack)))



;; declarative -> s
(defun-nondeterministic declarative (sent xlist-stack)
  (mvlet (((decl mask rem1 stack1)
	   (s sent xlist-stack)))
	 (returned-values (list 'decl decl)
		 mask
		 rem1
		 stack1)))

;; s -> subj verb verb-args verb-mods
(defun-nondeterministic s (phrase xlist-stack)
  (mvlet (((parsed-subj agmt type rem1 stack1)
	   (subj phrase xlist-stack)))
	 (mvlet (((parsed-verb type voice rem2 stack2)
		  (verb agmt rem1 stack1)))
		(format t "after verb~%")
		(mvlet (((args mask0 rem3 stack3)
			 (verb-args type voice (empty) rem2 stack2)))
		       (format t "after verb-args~%")
		       (mvlet (((mods set mask rem4 stack4)
				(verb-mods (minus (s-all) mask0)
					   (plus (s-all) mask0) '()
					   rem3 stack3)))
			      (format t "after verb-mods~%")
			      (returned-values (list 's parsed-subj 
						     (list 'vp parsed-verb 
							   args mods))
					       mask rem4 stack4))))))

;; subjects
(defun-nondeterministic subj (phrase xlist-stack)
  (either (subj1 phrase xlist-stack)
	  (subj2 phrase xlist-stack)
	  (subj3 phrase xlist-stack)))


(defun-nondeterministic subj1 (phrase xlist-stack)
  (mvlet (((term rem stack) (terminal phrase xlist-stack)))
	 (if (spec-eq? term 'there)
	     (returned-values (list 'existential term) '(_ _)
			      '(_ be) rem stack)
	     (fail))))

(defun-nondeterministic subj2 (phrase xlist-stack)
  (mvlet (((parsed-subj agmt s-case type mask rem stack) 
	   (np 'subj (s-all) phrase xlist-stack)))
	 (if (match? s-case (subj-case))
	     (returned-values parsed-subj agmt type rem stack)
	     (fail))))

(defun-nondeterministic subj3 (phrase xlist-stack)
  (mvlet (((parsed-np1 agmt1 s-case type1 mask1 rem1 stack1)
	   (np 'subj (s-all) phrase xlist-stack)))
	 (if (match? (subj-case) s-case)
	     (mvlet (((parsed-conj conj0 arg1 rem2 stack2) ;; ** check
		      (conj '() '() rem1 stack1))) ;; check left & right
		    (mvlet (((parsed-subj agmt2 type2 rem3 stack3)
			     (subj rem2 stack2)))
			   (if (and (match? agmt1 agmt2)
				    (match? type1 type2))
			       (returned-values (list 'np parsed-np1 (list 
								       parsed-conj
								       parsed-subj))
						agmt1 type1 rem3 stack3)
			       (fail))))
	     (fail))))

(defvar *np-count* 0)

;; nps
(defun-nondeterministic np (role-val set phrase xlist-stack)
  (either (np0 role-val set phrase xlist-stack)
	  (np1 role-val set phrase xlist-stack)
	  (np2 role-val set phrase xlist-stack)
	  (np3 role-val set phrase xlist-stack)
	  (np4 role-val set phrase xlist-stack)
	  (np5 role-val set phrase xlist-stack)))

;;;  This function will convert a given 0 <= num <= 9 to an asteriks 
;;;   delemited string
;;;  This function is added primarily for the detection of 'complex sentences'
;;;  in the Montague syntax conversion
;;;  the constant 48 <==> (char 0)
(defun num->str (num)
  (concatenate 'string 
	       "*" 
	       (make-string 1 :initial-element 
			    (int-char (+ num 48)))
	       "*"))
	
(defun str->num (str)
  (- (char-int (aref str 1)) 48))

(defun-nondeterministic np0 (role-val set phrase xlist-stack)
  (mvlet (((non-term stack) (virtual xlist-stack)))
	 (if (not (eq non-term 'np))
	     (fail)
	     (returned-values 
	       (list 'np `((np-num 
			    ,(num->str (locally-setf *np-count*
						     (+ 1 *np-count*))))
			    (trace (tied-to-np-num 
				     ,(num->str 
				       (lookup-np-count stack))))))
	       '(_ _) (vector '_ '_ '_)
	       '_ (empty) phrase stack))))

;;; This function is used to bind traces to their appropriate np's
(defun lookup-np-count (xlist-stack)
  (if (not (null xlist-stack))
      (if (listp (top-symbol xlist-stack))
	  (if (eq (car (top-symbol xlist-stack))
		  'close-paren)
	      (second (top-symbol xlist-stack))
	      (lookup-np-count (cdr xlist-stack)))
	  (lookup-np-count (cdr xlist-stack)))
      (fail)))


(defun-nondeterministic np1 (role-val set phrase xlist-stack)
  (mvlet (((parsed-pronoun gender agmt case rem stack) 
	   (pers-pron phrase xlist-stack)))
	 (mvlet (((case1 role1 np-case) (role)))
		(if (and (spec-eq? role1 'decl)
			 (spec-eq? case1 case)
			 (spec-eq? case role-val))
		    (returned-values (list 'np `((np-num 
						   ,(num->str
						      (locally-setf 
							*np-count*
							(+ 1 *np-count*))))
						  ,parsed-pronoun)
					   agmt gender)
			    agmt np-case 'def (empty) rem stack)
		    (fail)))))
			 
(defun-nondeterministic np2 (role-val set phrase xlist-stack)
  (mvlet (((kernel agmt type mods rem1 stack1)
	   (np-head phrase xlist-stack)))
	 (let ((np-count (locally-setf *np-count* (+ 1 *np-count*))))
	   (mvlet (((postmods mask rem2 stack2)
		    (np-compls type agmt role-val (np-all) 
			       rem1 stack1)))
			 (returned-values (list 'np `((np-num ,(num->str 
								np-count))
						,kernel) agmt 
						(append mods postmods))
					  agmt (vector '_ '_ '_) type mask 
					  rem2 stack2)))))

(defun-nondeterministic np3 (role-val set phrase xlist-stack)
  (mvlet (((parsed-det num type rem stack)
	   (determiner phrase xlist-stack)))
	 (mvlet (((symb rem2 stack2) (terminal rem stack)))
		(if (not (spec-eq? symb 'of))
		    (fail)
		    (mvlet (((parsed-np agmt case1 type mask rem1 stack1)
			     (np role-val (s-all) rem2 stack2)))
			   (if (and (match? agmt '(3 plu))
				    (match? case case1))
			       (returned-values 
				 `(np `((np-num 
					 ,(num->str (locally-setf
						      *np-count*
						      (+ 1 *np-count*))))
					(part ,det ,parsed-np)))
				 (list 3 num) (vector '_ '_ '_)
				 type mask rem1 stack1)
			       (fail)))))))

	 
;; np -> compound-pers-pron
(defun-nondeterministic np4 (role-val set phrase xlist-stack)
  (mvlet (((parsed-pronoun gender agmt case rem stack) 
	   (compound-pers-pron phrase xlist-stack)))
;	 (format t "parsed-pron:~s~%" parsed-pronoun)
	 (mvlet (((case1 role1 np-case) (role)))
		(if (and (spec-eq? role1 'decl)
			 (spec-eq? case1 case)
			 (spec-eq? case role-val))
		    (returned-values
		      (list 'np  
			    `((np-num 
				 ,(num->str (locally-setf *np-count*
							  (+ 1 *np-count*))))
			      (,parsed-pronoun
				(tied-to-np-num 
				,(num->str (- *np-count* 1)))))
			      ;;(,(nconc parsed-pronoun 
			;;	 `(tied-to-np-num 
				;;   ,(num->str (- *np-count* 1)))))
			    			    agmt gender)
		      agmt np-case 'def (empty) rem stack)
		    (fail)))))



(defun-nondeterministic np5 (role-val set phrase xlist-stack)
  (mvlet (((parsed-np agmt type mods rem1 stack1)
	   (np-head phrase xlist-stack)))
	 (let ((np-count (locally-setf *np-count* (+ 1 *np-count*))))
	   (mvlet (((parsed-pronoun gender agmt1 case rem2 stack2) 
		    (compound-pers-pron rem1 stack1)))
		  (mvlet (((case1 role1 np-case) (role)))
			 (if (and (spec-eq? role1 'decl)
				;;  (spec-eq? case1 case) case is 'obj
				;;  (spec-eq? case role-val)
				  (match? agmt agmt1))
			     (mvlet (((postmods mask rem3 stack3)
				      (np-compls type agmt role-val (np-all) 
						 rem2 stack2)))
				    (returned-values 
				      (list 'np 
					    `(((np-num ,(num->str np-count))
						,parsed-np)
					      ;;,(nconc )
					      (,parsed-pronoun
						(tied-to-np-number 
						,(num->str np-count))))
					    agmt mods)
				      agmt (vector '_ '_ '_)
				      type mask rem3 stack3))
			     (fail)))))))



(defun-nondeterministic np-head (phrase xlist-stack)
  (either (mvlet (((kernel0 agmt0 type0 rem1 stack1)
		   (np-head0 phrase xlist-stack)))
		 (mvlet (((kernel agmt type mods rem2 stack2)
			  (possessive kernel0 agmt0 type0 '() rem1 stack1)))
			(returned-values kernel agmt type mods rem2 stack2)))
	  (mvlet (((non-term stack) (virtual xlist-stack)))
		 (if (not (eq non-term 'np))
		     (fail)
		     (returned-values 
		       (list 'np-head `((np-num 
					 ,(num->str (locally-setf 
						      *np-count*
						      (+ 1 *np-count*))))
					(trace (tied-to-np-num 
						 ,(num->str (lookup-np-count
							     stack))))))
		       '(_ _) '(_ _) '() phrase stack)))))


(defun-nondeterministic np-compls (type agmt case set phrase xlist-stack)
  (either (np-compls1 type agmt case set phrase xlist-stack)
	  (np-compls2 type agmt case set phrase xlist-stack)))

(defun-nondeterministic np-compls1 (type agmt case set phrase xlist-stack)
  (if (match? type '(_ proper))
      (returned-values '() (empty) phrase xlist-stack)   
      (fail)))


(defun-nondeterministic np-compls2 (type agmt case set0 phrase xlist-stack)
  (if (match? type '(_ common))
      (mvlet (((rel-val1 mods set mask0 rem1 stack1) 
	       (np-mods agmt case '() set0 (np-all) phrase xlist-stack)))
	     (mvlet (((rel-val2 mask rem3 stack3)
		      (relative agmt set mask0 rem1 stack1)))
		    (returned-values (append mods rel-val2) mask rem3 stack3)))                     ;;** bug.. what is mods??
  (fail)))



(defun-nondeterministic np-head0 (phrase xlist-stack)
  (either (np-head00 phrase xlist-stack)
	  (np-head01 phrase xlist-stack)
	  (np-head02 phrase xlist-stack)
	  (np-head03 phrase xlist-stack)
	  (np-head04 phrase xlist-stack)))

(defun-nondeterministic np-head00 (phrase xlist-stack)
  (mvlet (((non-term stack)
	   (virtual xlist-stack)))
	 (if (eq non-term 'np-head00)
	     (returned-values '() '() '() phrase stack);; ** check return returned-values..
	     (fail))))

;; ** CHECK
;; np-head01 -> name
(defun-nondeterministic np-head01 (phrase xlist-stack)
  (mvlet (((parsed-name rem stack)
	   (name phrase xlist-stack)))
	 (returned-values (list 'name parsed-name)
		 '(3 sin)
		 '(def proper)
		 rem
		 stack)))

;; np-head02 -> determiner adjs noun     
(defun-nondeterministic np-head02 (phrase xlist-stack)
  (mvlet (((parsed-det num type rem1 stack1)
	   (determiner phrase xlist-stack))) 
	 (mvlet (((adjs-parse rem2 stack2)
		  (adjs rem1 stack1)))
		(mvlet (((noun-parse agmt rem3 stack3) 
			 (noun rem2 stack2)))
		       (returned-values (list 'np-head02 
					      parsed-det adjs-parse noun-parse)
					(list 3 num)
					(list 'def 'common)
					rem3
					stack3)))))


;; np-head03 -> poss-pron
(defun-nondeterministic np-head03 (phrase xlist-stack)
  (mvlet (((pronoun-val gender agmt rem stack)
	   (poss-pron phrase xlist-stack)))
	 (returned-values `(,pronoun-val ,gender) agmt '(def proper) rem 
			  (push-xlist 'nogap 'nonterminal 'gen-marker stack))))


;; np-head04 -> quantifier-pron
(defun-nondeterministic np-head04 (phrase xlist-stack)
  (mvlet (((parsed-det noun-val rem stack) 
	   (quantifier-pron phrase xlist-stack)))
	 (returned-values (list 'np-head04 parsed-det '() noun-val)
		 (list 3 'sin)
		 (list 'indef 'common)
		 rem stack)))




(defun-nondeterministic possessive (kernel agmt type mods-so-far phrase xlist-stack)
  (either (possessive1 kernel agmt type mods-so-far phrase xlist-stack)
	  (possessive2 kernel agmt type mods-so-far phrase xlist-stack)))

(defun-nondeterministic possessive1 (kernel0 agmt0 type0 mods-so-far phrase xlist-stack)
  (mvlet (((rem1 stack1) 
	   (gen-case phrase xlist-stack)))
	 (mvlet (((kernel1 agmt1 type1 rem2 stack2)
		  (np-head0 rem1 stack1)))
		(mvlet (((kernel agmt type mods phrase xlist-stack)  
			 (possessive kernel1 agmt1 
				     type1 
				     (list 'pp 
					   (list 'poss 
						 (list 'np agmt0 
						       kernel0 mods-so-far)))
				     rem2 stack2)))
		       (returned-values kernel agmt type mods phrase xlist-stack)))))

(defun-nondeterministic possessive2 (kernel agmt type mods phrase xlist-stack)
  (returned-values kernel agmt type mods phrase xlist-stack))


(defun-nondeterministic gen-case (phrase xlist-stack)
  (mvlet (((rem stack) 
	   (gen-marker phrase xlist-stack)))
	 (returned-values rem (push-xlist 'nogap 'terminal 'the stack))))

(defun-nondeterministic gen-marker (phrase stack)
  (either (gen-marker1 phrase stack)
	  (gen-marker2 phrase stack)))

(defun-nondeterministic gen-marker1 (phrase stack)
  (mvlet (((non-term stack1) (virtual stack)))
	 (if (eq non-term 'gen-marker)
	     (returned-values phrase stack1)
	     (fail))))

(defun-nondeterministic gen-marker2 (phrase stack)
  (mvlet (((term rem1 stack1)  
	   (terminal phrase stack)))
	 (if (equal ''s term);; include cases like chess'
	     ;(mvlet (((rem2 stack2)
	     ;;  (an-s rem stack)))
		   ; )
	     (returned-values rem1 stack1)
	     (fail))))

;; this function is combined w/gen-marker2
(defun-nondeterministic an-s (phrase xlist-stack)
  (either (mvlet (((rem stack)
		   (terminal 's phrase xlist-stack)))
		 (returned-values rem stack))
	  (returned-values phrase xlist-stack)))

(defun-nondeterministic determiner (phrase xlist-stack)
  (either (mvlet (((parsed-det num type rem stack)
		   (det phrase xlist-stack)))
		 (returned-values parsed-det num type rem stack))
	  (mvlet (((parsed-det num type rem stack)
		   (quant-phrase phrase xlist-stack)))
		 (returned-values parsed-det num type rem stack))))
		
;; ** BUG num is undefined!	       
(defun-nondeterministic quant-phrase (phrase xlist-stack)
  (mvlet (((op def rem1 stack1) (quant phrase xlist-stack)))
;;	 (mvlet (((parsed-quant rem2 stack2) (number num rem1 stack1))))
	 (returned-values (list 'quant op) '_ ; num
			  def rem1 stack1)))


(defun-nondeterministic quant (phrase xlist-stack)
  (either (quant1 phrase xlist-stack)
	  (quant2 phrase xlist-stack)
	  (quant3 phrase xlist-stack)
	  (quant4 phrase xlist-stack)))
		 
(defun-nondeterministic quant1 (phrase xlist-stack)      
  (mvlet (((op rem1 stack1)
	   (neg-adv phrase xlist-stack)))
	 (mvlet (((op rem2 stack2)
		  (comp-adv rem1 stack1)))
		(mvlet (((term rem3 stack3)
			 (terminal rem2 stack2)))
		       (if (not (eq 'than term))
			   (fail)
			   (returned-values op 'indef rem3 stack3))))))

(defun-nondeterministic quant2 (phrase xlist-stack)
  (mvlet (((term rem1 stack1)
	   (terminal phrase xlist-stack)))
	 (if (not (eq 'at term))
	     (fail)
	     (mvlet (((sup-adv-val rem2 stack2)
		      (sup-adv rem1 stack1)))
		    (mvlet (((op rem3 stack3)
			     (sup-op sup-adv-val rem2 stack2)))
			   (returned-values op 'indef rem3 stack3))))))


(defun-nondeterministic quant3 (phrase xlist-stack)
  (mvlet (((term rem stack)
	   (terminal phrase xlist-stack)))
	 (if (not (eq term 'the))
	     (fail)
	     (returned-values 'the 'indef rem stack))))

(defun-nondeterministic quant4 (phrase xlist-stack)
  (returned-values 'same 'indef phrase xlist-stack))

(defun-nondeterministic neg-adv (phrase xlist-stack)
  (either (mvlet (((term rem stack)
		   (terminal phrase xlist-stack)))
		 (if (not (eq 'not term))
		     (fail)
		     (returned-values '(not) rem stack)));; ** bug adv? (null) is it i/o??
	  (returned-values  '() phrase xlist-stack))) ; ** bug adv? (null) is it i/o??



;(defun-nondeterministic sup-op (adv phrase xlist-stack)
 ; (either (if (spec-eq? adv 'least)
	;      (returned-values (list '(not adv)) phrase xlist-stack)
	 ;     (fail))
	  ;(if (spec-eq? adv 'most)
	 ;     (returned-values (list '(not adv)) phrase xlist-stack)
	  ;    (fail))))



(defun-nondeterministic np-mods (agmt case mods-so-far set0 mask1 phrase xlist-stack)
  (either (mvlet (((mod mask0 rem1 stack1) 
		   (np-mod agmt case set0 phrase xlist-stack)))
		 (mvlet (((rel-val mods set mask rem2 stack2)
			  (np-mods agmt case (append mod mods-so-far)
				   (minus set0 (plus (trace1) mask0))
				   (plus mask0 set0) rem1 stack1)))
			(returned-values rel-val mods set mask rem2 stack2)))
	  (returned-values mods-so-far mods-so-far set0 mask1 phrase xlist-stack)))

				   
(defun-nondeterministic np-mod (agmt case set phrase xlist-stack)
  (either (mvlet (((parsed-pp mask rem stack)
		   (prep-p case set phrase xlist-stack)))
		 (returned-values parsed-pp mask rem stack))
	       agmt case set phrase xlist-stack)
	  (mvlet (((agmt1 parsed-wh mask rem stack)
		   (reduced-relative set phrase xlist-stack)))
		 (if (match? agmt agmt1)
		     (returned-values parsed-wh mask rem stack)
		     (fail))))


(defun-nondeterministic prep-p (case set phrase stack)
  (mvlet (((parsed-prep rem1 stack1)
	   (prep phrase stack)))
	 (mvlet (((parsed-np agmt p-case type mask rem2 stack2)
		  (np case set rem1 stack1)))
		(if (match? (prep-case) p-case)
		    (returned-values 
		      (list 'pp parsed-prep 
			    parsed-np)
		      mask rem2 stack2)
		    (fail)))))


;;  Reduced Relative clause
(defun-nondeterministic reduced-relative (set phrase xlist-stack)
  (if (match? set (is-pred))
      (mvlet (((agmt conj-val rel-val mask rem stack)
	       (reduced-rel-conj phrase xlist-stack)))
	     (returned-values agmt rel-val mask rem stack))
      (fail)))


(defun-nondeterministic reduced-rel-conj (phrase xlist-stack)
  (mvlet (((agmt rel0 mask0 rem1 stack1)
	   (reduced-rel phrase xlist-stack)))
	 (mvlet (((agmt conj-val rel-val mask rem2 stack2)
		  (reduced-rel-rest rel0 mask0 rem1 stack1)))
		(returned-values agmt conj-val rel-val mask rem2 stack2))))

(defun-nondeterministic reduced-rel-rest (rel0 mask0 phrase xlist-stack)
  (either (mvlet (((conj0 conj-val rel-val rem1 stack1)
		   (conj rel0 '() phrase xlist-stack)))
		 (mvlet (((agmt conj-val1 rel11 mask rem2 stack2)
			  (reduced-rel-conj rem1 stack1)))
			(if (match? conj-val conj-val1)
			    (returned-values agmt conj0 rel-val mask
					     rem2 stack2)
			    (fail))))
	  (returned-values '(_ _) '() rel0 mask0 phrase xlist-stack)))

(defun-nondeterministic reduced-rel (phrase xlist-stack)
  (mvlet (((rem1 stack1)
	   (open-paren phrase xlist-stack)))
	 (mvlet (((agmt x rem2 stack2)
		  (reduced-wh rem1 stack1)))
		(mvlet (((parsed-s mask0 rem3 stack3)
			 (s rem2 stack2)))
		       (mvlet (((rem4 stack4) 
				(close-paren rem3 stack3)))
			      (returned-values agmt 
				(list 'reduced-rel `(,x 
						      ,parsed-s))
				(minus mask0 (trace1)) rem4 stack4))))))  

(defun-nondeterministic reduced-wh (phrase xlist-stack)
  (either (reduced-wh1 phrase xlist-stack)
	  (reduced-wh2 phrase xlist-stack)
	  (reduced-wh3 phrase xlist-stack)))

;; (mvlet 
;;     (((set0 mask0)
;;       (trace0)))
;;     (mvlet 
;;       (((parsed-np agmt np-case type mask00 rem1 stack1)
;; 	(np '_ set0 phrase xlist-stack)))
;;       (if (and (match? mask0 mask00)
;; 	       (match? (subj-case) np-case))
;; 	  (mvlet 
;; 	    (((root-val parsed-verb tense type rem2 stack2)
;; 	      (verb-form agmt rem1 stack1)))
;; 	    (if (and (spec-eq? root-val 'be)
;; 		     (match? tense '(pres fin))
;; 		     (match? type '(main _)))
;; 		(mvlet 
;; 		  (((neg-val rem3 stack3)
;; 		    (neg '(_ _) rem2 stack2)))
;; 		  (mvlet 
;; 		    (((pred-val mask rem4 stack4)
;; 		      (pred 
;; 			(if (listp neg-val)
;; 			    (car neg-val)
;; 			    neg-val)
;; 			rem3 stack3)))
;; )))))))


(defun-nondeterministic reduced-wh1 (phrase xlist-stack)
  (mvlet (((neg-val rem1 stack1)
	   (neg '(_ _) phrase xlist-stack)))
	 (mvlet (((pred-val mask1 rem2 stack2)
		  (pred 
		    (if (listp neg-val)
			(car neg-val)
			neg-val)
		    rem1 stack1)))
		(returned-values 
		  '(_ _) `(x '()) rem2
		  (push-xlist 'nogap 'nonterminal 'np 
			      (push-xlist 'nogap 'nonterminal 'verb-form 
					  (push-xlist 'nogap 'nonterminal 'neg
						      (push-xlist 'nogap 
								  'nonterminal
								  'pred
								  stack2))))))))


;; (mvlet (((set mask)
;; 	   (trace0)))
;; 	 (mvlet (((parsed-np agmt np-case type mask0 rem1 stack1)
;; 		  (np '_ set phrase xlist-stack)))
;; 		(if (and (match? mask mask0)
;; 			 (match? (subj-case) np-case))
;; 		    (mvlet (((parsed-verb type voice rem2 stack2)
;; 			     (verb agmt rem1 stack1)))
;; ))))
;; (defun get-root (parsed-verb)
;;   (second (third parsed-verb)))	
	   
(defun-nondeterministic reduced-wh2 (phrase xlist-stack)
  (mvlet (((parsed-participle type1 voice1 rem stack)
	   (participle phrase xlist-stack)))
	 (returned-values '(_ _) `(x '()) rem 
			  (push-xlist 'nogap 'nonterminal 'np
				      (push-xlist 'nogap 'nonterminal 'verb 
						  stack)))))
			      

;; (mvlet (((set mask)
;; 	   (trace0)))
;; 	 (mvlet (((parsed-subj agmt0 s-case type mask0 rem1 stack1)
;; 		  (np '_ set phrase xlist-stack)))
;; 		(if (and (match? mask mask0)
;; 			 (match? (subj-case) s-case)
;; 			 (match? type '(def _)))	
;; 		    (mvlet (((parsed-np agmtX v-case type1 mask1 rem2 stack2)
;; 			     (np '_ set rem1 stack1)))
;; 			   (if (and (match? mask mask1)
;; 				    (match? (verb-case) v-case))
;; 			       )))))


;;; BUG (all-values (np 'subj (s-all) '(the room) '((GAP NONTERMINAL CLOSE-PAREN))))

(defun-nondeterministic reduced-wh3 (phrase xlist-stack)
  (mvlet (((parsed-subj agmt case type mask rem stack)
	   (np 'subj (s-all) phrase xlist-stack)))
	 (if (match? type '(def _))
	     (returned-values agmt `(x '()) rem 
			      (push-xlist 'gap 'nonterminal 'np
					  (push-xlist 'nongap 'nonterminal
						      stack)))
	     (fail))))


(defun-nondeterministic relative (agmt set mask0 phrase xlist-stack)
  (either (if (match? set (is-pred))
	      (mvlet (((conj-val0 rel-val mask rem stack)
		       (rel-conj agmt phrase xlist-stack)))
		     (returned-values `(,conj-val0 ,rel-val) mask rem stack))
	      (fail))
	  (returned-values '() mask0 phrase xlist-stack)))

(defun-nondeterministic rel-conj (agmt phrase xlist-stack)
  (mvlet (((rel0 mask0 rem1 stack1)
	   (rel agmt phrase xlist-stack)))
      (mvlet (((agmt1 conj-val0 rel2 mask2 rem2 stack2)
	       (rel-rest agmt rel0 mask0 rem1 stack1)))
	     (if (match? agmt agmt1)
		 (returned-values conj-val0 rel2 mask2 rem2 stack2)
		 (fail)))))


;; ** check
(defun-nondeterministic rel-rest (agmt rel0 mask0 phrase xlist-stack)
  (either (mvlet (((conj0 conj-val rel-val rem1 stack1)
		   (conj rel0 '() phrase xlist-stack)))
		 (mvlet (((conj-val0 rel-val1 mask rem2 stack2)
			  (rel-conj agmt rem1 stack1)))
			(returned-values agmt rel0 `(,conj0 ,rel-val1)
					 mask rem2 stack2)))
	  (returned-values '(_ _) '() rel0 mask0 phrase xlist-stack)))

(defun-nondeterministic rel (agmt phrase xlist-stack)
  (mvlet (((rem1 stack1)
	   (open-paren phrase xlist-stack)))
	 (mvlet (((x rem2 stack2)
		  (variable agmt rem1 stack1)))
		(mvlet (((parsed-s mask0 rem3 stack3)
			 (s rem2 stack2)))
		       (mvlet (((rem4 stack4) 
				(close-paren rem3 stack3)))
			      (returned-values 
				(list 'rel `(,x ,parsed-s))
				(minus mask0 (trace1)) rem4 stack4))))))

(defun-nondeterministic variable (agmt phrase xlist-stack)
  (either
    (mvlet (((term rem2 stack2) 
	     (terminal phrase xlist-stack)))		       
	   (if (eq term 'that)
	       (mvlet (((set0 mask0)
			(trace0)))
		      (returned-values 
			(list 'wh (list 'that))  
			rem2 
			(push-xlist 'gap 
				    'nonterminal 'np
			;;;		`(np `(np `(,agmt wh `(,x) ())))
				    stack2)))
	       (fail)))
    (mvlet (((x agmt parsed-np np-case rem1 stack1)
	     (wh phrase xlist-stack)))
	   (mvlet (((set0 mask0)
		    (trace0)))
		  (returned-values 
		    agmt (list  'wh x) rem1
		    (push-xlist 'gap 'nonterminal 'np stack1))))
    (mvlet (((parsed-prep rem1 stack1)
	     (prep phrase xlist-stack)))
	   (mvlet (((x agmt parsed-pp np-case rem1 stack1)
		    (wh rem1 stack1)))
		  (mvlet (((set0 mask0)
			   (trace0)))
			 (if (match? np-case (compl-case))
			     (returned-values 
			       agmt (list  'wh x) rem1
			       (push-xlist 'gap 'nonterminal 'pp stack1))
			     (fail)))))))


;; (defun-nondeterministic variable (agmt phrase xlist-stack)
;;   (either
;;     (mvlet (((term rem2 stack2) 
;; 	     (terminal phrase xlist-stack)))		       
;; 	   (if (eq term 'that)
;; 	       (mvlet (((set0 mask0)
;; 			(trace0)))
;; 		      (if (and (match? set0 set)
;; 			       (match? mask0 mask))
;; 			  (returned-values 
;; 			    (list 'wh (list 'that))  
;; 			    rem2 
;; 			    (push-xlist 'gap 
;; 					'nonterminal 'np
;; 			;;;		`(np `(np `(,agmt wh `(,x) ())))
;; 					stack2))
;; 			  (fail)))
;; 	       (fail)))
;;     (mvlet (((x agmt parsed-np np-case rem1 stack1)
;; 	     (wh phrase xlist-stack)))
;; 	   (mvlet (((set0 mask0)
;; 		    (trace0)))
;; 		  (if (and (match? set0 set)
;; 			   (match? mask0 mask))
;; 		      (returned-values 
;; 			agmt (list  'wh x) rem1
;; 			(push-xlist 'gap 'nonterminal 'np stack1))
;; 		      (fail))))
;;     (mvlet (((parsed-prep rem1 stack1)
;; 	     (prep phrase xlist-stack)))
;; 	   (mvlet (((x agmt parsed-pp np-case rem1 stack1)
;; 		    (wh rem1 stack1)))
;; 		  (mvlet (((set0 mask0)
;; 			   (trace0)))
;; 			 (if (and (match? set0 set)
;; 				  (match? mask0 mask)
;; 				  (match? case (compl-case)))
;; 			     (returned-values 
;; 			       agmt (list  'wh x) rem1
;; 			       (push-xlist 'gap 'nonterminal 'pp stack1))
;; 			     (fail)))))))

		  
(defun-nondeterministic wh (phrase xlist-stack)
  (either (mvlet (((parsed-case rem1 stack1)
		   (rel-pron phrase xlist-stack)))
		 (mvlet (((case type np-case)
			  (role)))
			(if (and (spec-eq? parsed-case case)
				 (spec-eq? 'decl type))
			    (returned-values agmt '(x) ;; empty X
					     `(np ,agmt `(wh `(x)) `())
					     np-case rem1 stack1)
			    (fail))))
	  (mvlet (((kernel0 agmt type0 rem1 stack1)
		   (np-head0 phrase xlist-stack)))
		 (if (match? type0 '(_ common))
		     (mvlet (((parsed-prep rem2 stack2)
			      (prep rem1 stack1)))
			    (mvlet (((x agmt0 parsed-np np-case rem3 stack3)
				     (wh rem2 stack2)))
				   (returned-values x agmt0
						    `(np ,agmt kernel0 
							 `(pp `(,parsed-prep
								 ,parsed-np)))
						    rem3 stack3)))
		     (fail)))
	  (mvlet (((x agmt0 rem1 stack1)
		   (whose phrase xlist-stack)))
		 (mvlet (((parsed-np agmt case type mask rem2 stack2)
			  (np 'subj (s-all) rem1 stack1)))
			(if (match? type '(def _))
			    (returned-values x agmt0 parsed-np case 
					     rem2 stack2)
			    (fail))))))

(defun-nondeterministic whose (phrase xlist-stack)
  (mvlet (((term rem stack)
	   (terminal phrase xlist-stack)))
	 (if (eq term 'whose)
	     (returned-values (push-xlist 'npgap 'nonterminal
					  'np-head0
					  (push-xlist 'nogap 
						      'nonterminal 'gen-marker
						      ;; `(np-head0 `(wh `(x)))
						      stack)))
	     (fail))))

(defun-nondeterministic open-paren (phrase stack)
  (returned-values phrase (push-xlist 'gap 'nonterminal 
				      `(close-paren ,*np-count*) stack)))

(defun-nondeterministic close-paren (phrase stack)
  (mvlet (((non-term stack1) 
	   (virtual stack)))
	 (if (eq (car non-term) 'close-paren)
	     (returned-values phrase stack1)
	     (fail))))


;;; ** CHECk maybe set should NOT be returned!!
;; verb modifiers..
(defun-nondeterministic verb-mods (set0 mask1 mods-so-far phrase xlist-stack)
  (either (mvlet (((mod mask0 rem1 stack1)
		   (verb-mod set0 mask1 phrase xlist-stack)))
		 (mvlet (((mods set mask rem2 stack2)
			  (verb-mods (minus set0 (plus (trace1) mask0))
				     (plus mask0 set0) 
				     (append mod mods-so-far) 
				     rem1 stack1)))
			(returned-values mods set mask rem2 stack2)))
	  (returned-values mods-so-far set0 mask1 phrase xlist-stack)))


(defun-nondeterministic verb-mod (set mask phrase xlist-stack)
  (either (mvlet (((adv-val mask0 rem stack)
		   (adv-phrase set phrase xlist-stack)))
		 (returned-values adv-val mask rem stack))
	  (if (and (match? mask (empty))
		   (match? set (is-adv)))
	      (mvlet (((adverb-val rem stack)
		       (adverb phrase xlist-stack)))
		     (returned-values adverb-val mask rem stack))
	      (fail))
	  (mvlet (((parsed-pp mask1 rem stack)
		   (prep-p 'compl set phrase xlist-stack)))
		 (returned-values parsed-pp mask1 rem stack))))
		 
	      


;; verb phrases..
;; note that this might be a bug in the paper!  Role should not be i/o to 
;; verb_form..  remove it from this implementation

(defun-nondeterministic verb (agmt phrase xlist-stack)
  (either (verb1 agmt phrase xlist-stack)
	  (verb1 agmt phrase xlist-stack)))

(defun-nondeterministic verb1 (agmt phrase xlist-stack)
  (mvlet (((root parsed-verb tense type0 rem1 stack1)
	   (verb-form agmt phrase xlist-stack)))
	 (mvlet (((neg-val rem2 stack2)
		  (neg type0 rem1 stack1)))
		(let ((role-val (car type0)))
		  (mvlet (((parsed-rest-verb voice aspect rem3 stack3) 
			   (rest-verb role-val root agmt rem2 stack2)))
			 role-val root rem2)
		  (format t "VERB: type0:~s voice:~s rem2:~s rem3:~s stack3:~s"
			  type0 voice rem2 rem3 stack3)
		  (returned-values
		    (list 'v1 parsed-verb 
			  (if (eq root (second parsed-verb))
			      (list 'root 'same_as_verb)
			      (list 'root root))
			  voice tense aspect 
			  `(,neg-val ,parsed-rest-verb))
		    type0 voice rem3 stack3)))))


;;  this next clause does not accomodate for passive **check

(defun-nondeterministic verb2 (agmt phrase xlist-stack)
  (mvlet (((root parsed-verb tense type0 rem1 stack1)
	   (verb-form agmt phrase xlist-stack)))
	 (mvlet (((neg-val rem2 stack2)
		  (neg type0 rem1 stack1)))
		(format t "VERB: type0:~s voice:~s rem2:~s stack2:~s"
			  type0 'active rem2 stack2)
		(let ((verb-val (list 'v parsed-verb 
			(if (eq root (second parsed-verb))
			    (list 'root 'same_as_verb)
			    (list 'root root))
			'active tense '() 
			`(,neg-val)))) 
		(returned-values
		  verb-val
		  type0 'active rem2 stack2)))))

;; (defun-nondeterministic verb (agmt phrase xlist-stack)
;;   (mvlet (((root parsed-verb tense type0 rem1 stack1)
;; 	   (verb-form agmt phrase xlist-stack)))
;; 	 (mvlet (((neg-val rem2 stack2)
;; 		  (neg type0 rem1 stack1)))
;;                 (either 
;; 		  (let ((role-val (car type0)))
;; 		    (mvlet (((parsed-rest-verb voice aspect rem3 stack3) 
;; 			     (rest-verb role-val root agmt rem2 stack2)))
;; 				   role-val root rem2)
;; 	   (format t "VERB: type0:~s voice:~s rem2:~s rem3:~s stack3:~s"
;; 			   type0 voice rem2 rem3 stack3)
;; 			   (returned-values
;; 			     (list 'v1 parsed-verb 
;; 				   (if (eq root (second parsed-verb))
;; 					(list 'root 'same_as_verb)
;; 					(list 'root root))
;; 				   voice tense aspect 
;; 				   `(,neg-val ,parsed-rest-verb))
;; 			     type0 voice rem3 stack3))
;; 		;;;  this next clause does not accomodate for passive **check
;; 		  (returned-values
;; 		    (list 'v parsed-verb 
;; 			  (if (eq root (second parsed-verb))
;; 			      (list 'root 'same_as_verb)
;; 			      (list 'root root))
;; 			  'active tense '() 
;; 			  `(,neg-val))
;; 		    type0 'active rem2 stack2)))))



(defun-nondeterministic neg (type phrase stack)
  (either (if (match? type '(aux _))
	      (mvlet (((term rem1 stack1)
		       (terminal phrase stack)))
		     (if (eq term 'not)
			 (returned-values (list 'neg term) rem1 stack1)
			 (fail)))
	      (fail))
	  (returned-values 'pos phrase stack)))


(defun-nondeterministic rest-verb (role root agmt phrase xlist-stack)
  (either (rest-verb1 role root agmt phrase xlist-stack)
	  (rest-verb2 role root agmt phrase xlist-stack)
	  (rest-verb3 role root agmt phrase xlist-stack)
	  (rest-verb4 role root agmt phrase xlist-stack)
	  (rest-verb5 role root agmt phrase xlist-stack)
	  (rest-verb6 role root agmt phrase xlist-stack)
	  (rest-verb7 role root agmt phrase xlist-stack)
	  (rest-verb8 role root agmt phrase xlist-stack)
	  (rest-verb9 role root agmt phrase xlist-stack)))

  

(defun-nondeterministic rest-verb1 (role root agmt phrase xlist-stack)
  (if (eq role 'aux)
      (mvlet (((root-val parsed-verb tense type rem1 stack1)
	       (verb-form agmt phrase xlist-stack)))
	     (if (and (eq root-val 'have)
		      (match? tense '(past part)))
		 (mvlet (((root0 voice aspect rem2 stack2)
			  (have root-val agmt rem1 stack1)))
			(returned-values `(,parsed-verb ,root0)
					 voice `(perf ,aspect) rem2 stack2))
		 (fail)))
      (fail)))


(defun-nondeterministic rest-verb2 (role root agmt phrase xlist-stack)
  (if (eq role 'aux)
      (mvlet (((root-val parsed-verb tense type rem1 stack1)
	       (verb-form agmt phrase xlist-stack)))
	     (if (eq root-val 'be)
		 (mvlet (((tense0 root0 voice aspect rem2 stack2)
			  (be root-val agmt rem1 stack1)))
			(returned-values
			  (list parsed-verb 
				(list 'root 
				      (if (eq root-val (second parsed-verb))
					  (list 'root 'same_as_verb)
					  (list 'root root-val))) root0)
					 voice aspect rem2 stack2))
		 (fail)))
      (fail)))

(defun-nondeterministic rest-verb3 (role root agmt phrase xlist-stack)
  (if (and (eq role 'aux)
	   (eq root 'be))
      (mvlet (((tense0 root0 voice aspect rem1 stack1)
	       (be root agmt phrase xlist-stack)))
	     (if (eq 'active voice)
		 (returned-values root0 voice aspect rem1 stack1)
		 (fail)))
      (fail)))

(defun-nondeterministic rest-verb4 (role root agmt phrase xlist-stack)
  (if (eq role 'aux)
      (mvlet (((root-val parsed-verb tense type rem1 stack1)
	       (verb-form agmt phrase xlist-stack)))
	     (if (and (match? tense '(_ inf))
		      (eq root-val 'do))
		 (returned-values (list parsed-verb 
					(if (eq root-val (second parsed-verb))
					    (list 'root 'same_as_verb)
					    (list 'root root-val)))
				  'active '() rem1 stack1)
		 (fail)))
      (fail)))


(defun-nondeterministic rest-verb5 (role root agmt phrase xlist-stack)
  (if (not (spec-eq? role 'main))
      (mvlet (((root-val parsed-verb tense type rem stack)
	       (verb-form agmt phrase xlist-stack)))
	     (returned-values (list parsed-verb 
				    (if (eq root-val (second parsed-verb))
					(list 'root 'same_as_verb)
					(list 'root root-val))
				    'active '() rem stack) ))
      (fail)))



(defun-nondeterministic rest-verb6 (role root agmt phrase xlist-stack)
  (if (spec-eq? role 'main)
      (returned-values '() 'active '() phrase xlist-stack)
      (fail)))

(defun-nondeterministic rest-verb7 (role root agmt phrase xlist-stack)
  (mvlet (((root-val parsed-verb tense type rem stack)
	   (verb-form '(_ _) phrase xlist-stack)))
	 (if (eq root-val (second parsed-verb))
	     (returned-values (list '(simple future) parsed-verb 
				    (if (eq root-val (second parsed-verb))
					(list 'root 'same_as_verb)
					(list 'root root-val)))
				    'active '() rem stack)
	     (fail))))

;;; passive voice
(defun-nondeterministic rest-verb8 (role root agmt phrase xlist-stack)
  (if (eq role 'aux)
      (mvlet (((root-val parsed-verb tense type rem1 stack1)
	       (verb-form agmt phrase xlist-stack)))
	     (mvlet (((tense0 root0 voice aspect rem2 stack2)
		      (be root-val agmt rem1 stack1)))
		    (if (match? '(past part) tense0)
			(returned-values
			  (list parsed-verb 
				(if (eq root-val (second parsed-verb))
				    (list 'root 'same_as_verb)
				    (list 'root root-val)))
			  voice aspect rem2 stack2)
			(fail))))
      (fail)))

;; verb to inf-verb (likes to eat)
(defun-nondeterministic rest-verb9 (role root agmt phrase xlist-stack)
  (if (eq role 'main)
      (mvlet (((term-prep rem1 stack1)
	       (terminal phrase xlist-stack)))
	     (if (eq term-prep 'to)
		 (mvlet (((root-val parsed-verb tense type rem2 stack2)
			  (verb-form '(_ _) rem1 stack1)))
			(if (eq root-val (second parsed-verb))
			    (returned-values (list `(prep ,term-prep)
						   parsed-verb 
						   `(root same_as_verb))
					     'active '() rem2 stack2)
			    (fail)))
		 (fail)))
      (fail)))


(defun-nondeterministic have (root agmt phrase xlist-stack)
  (either (if (eq root 'be)
	      (mvlet (((root-val parsed-verb tense type  rem1 stack1)
		       (verb-form agmt phrase xlist-stack)))
		     (mvlet (((tense0 root0 voice aspect rem2 stack2)
			      (be root agmt rem1 stack1)))
			    (returned-values root0 voice aspect rem2 stack2)))
	      (fail))
	  (returned-values root 'active '() phrase xlist-stack)))

(defun-nondeterministic be (root agmt phrase xlist-stack)
  (either (returned-values '(past part) root 'passive '() phrase xlist-stack)
	  (mvlet (((root0 voice rem stack)
		   (passive root agmt phrase xlist-stack)))
		 (returned-values '(pres part) 
				  root0 voice '(prog) rem stack))))
                                              ;; ^--|  what is prog?! check **


(defun-nondeterministic passive (root agmt phrase xlist-stack)
  (either (if (eq root 'be)
	      (mvlet (((root-val parsed-verb tense type rem stack)
		       (verb-form agmt phrase xlist-stack)))
		     (if (and (match? type (passive-val))
			      (match? tense '(past part))
			      (not (member 'cont tense)))
			 (returned-values (list parsed-verb (list 'root 
								  root-val))
					  'passive rem stack)
			 (fail)))
	      (fail))
	  (if (eq root 'be)
	      (mvlet (((root-val parsed-verb tense type rem stack)
		       (verb-form agmt phrase xlist-stack)))
		     (if (and (match? type (passive-val))
			      (match? tense '(past part))
			      (member 'cont tense))
			 (returned-values (list parsed-verb (list 'root 
								  root-val))
					  'active rem stack)
			 (fail)))
	      (fail))
	  (returned-values root 'active phrase xlist-stack)))
		     
(defun-nondeterministic passive-val ()
  (either '(_ trans)
	  '(_ ditrans)))
	  

(defun-nondeterministic participle (phrase xlist-stack)
  (mvlet (((neg-val rem1 stack1)
	   (neg '(_ _) phrase xlist-stack)))
	 (mvlet (((root-val parsed-verb tense type rem2 stack2)
		  (verb-form '(_ _) rem1 stack1)))
		(mvlet (((participle-tense participle-voice participle-aspect)
			 (participle-val)))
		       (if (match? participle-tense tense)
			   (returned-values 
			     `(verb `(,parsed-verb `(root ,root-val))
				    participle-voice participle-aspect neg-val)
			     type participle-voice)
			   (fail))))))

(defun-nondeterministic participle-val ()
  (either (returned-values '(pres part) 'active '(prog))
	  (returned-values '(past part) 'passive '())))


;; verb args..
(defun-nondeterministic verb-args (type voice mask0 phrase xlist-stack)
  (either (verb-args1 type voice mask0 phrase xlist-stack)
	  (verb-args2 type voice mask0 phrase xlist-stack)
	  (verb-args3 type voice mask0 phrase xlist-stack)
	  (verb-args4 type voice mask0 phrase xlist-stack)
	  (verb-args5 type voice mask0 phrase xlist-stack)
	  (verb-args6 type voice mask0 phrase xlist-stack)
	  (verb-args7 type voice mask0 phrase xlist-stack)
	  (verb-args8 type voice mask0 phrase xlist-stack)
	  (verb-args9 type voice mask0 phrase xlist-stack)))


;;; ** CHECK  this is a quick and dirty hack to avoid infinite loops
(defun-nondeterministic verb-args0 (type voice mask0 phrase xlist-stack)
  (either (verb-args2 type voice mask0 phrase xlist-stack)
	  (verb-args3 type voice mask0 phrase xlist-stack)
	  (verb-args4 type voice mask0 phrase xlist-stack)
	  (verb-args5 type voice mask0 phrase xlist-stack)
	  (verb-args6 type voice mask0 phrase xlist-stack)
	  (verb-args7 type voice mask0 phrase xlist-stack)
	  (verb-args8 type voice mask0 phrase xlist-stack)
	  (verb-args9 type voice mask0 phrase xlist-stack)))

;;; ** check
(defun-nondeterministic verb-args1 (type voice mask0 phrase xlist-stack)
  (mvlet (((adv-args args rem1 stack1)
	   (advs '() (vector '_ '_ '_ '_) phrase xlist-stack)))
	 (mvlet (((parsed-args mask rem2 stack2)
		  (verb-args0 type voice mask0 rem1 stack1)))
		(returned-values (list 'verb_args adv-args parsed-args)
				 mask rem2 stack2))))
	       

(defun-nondeterministic verb-args2 (type voice mask0 phrase xlist-stack)
  (if (and (match? type '(_ trans))
	   (eq 'active voice))
      (mvlet (((parsed-np mask rem stack) 
	       (verb-arg phrase xlist-stack)))
	     (returned-values (list 'arg (list 'dir parsed-np)) 
			      mask rem stack))
      (fail)))


(defun-nondeterministic verb-args3 (type voice mask0 phrase xlist-stack)
  (if (match? type '(_ ditrans))
      (mvlet (((parsed-np mask1 rem1 stack1) 
	       (verb-arg phrase xlist-stack)))
	     (mvlet (((case-val parsed-obj rem2 stack2)
		      (object mask1 rem1 stack1)))
		    (returned-values (list 'verb_args (list case-val parsed-np)
					   parsed-obj)
	       		     mask1 rem2 stack2)))
      (fail)))


(defun-nondeterministic verb-args4 (type voice mask0 phrase xlist-stack)
  (if (match? type '(_ be))
      (mvlet (((term rem stack)
	       (terminal phrase xlist-stack)))
	     (if (eq term 'there)
		 (returned-values (list 'verb-args '(there)) mask0 rem stack)
		 (fail)))
      (fail)))

;;;; ** check
(defun-nondeterministic verb-args5 (type voice mask0 phrase xlist-stack)
  (if (match? type '(_ be))
      (mvlet (((conj-val pred-val mask rem stack)
	       (pred-conj phrase xlist-stack)))
	    (returned-values (list 'verb-args (list 'pred pred-val))
			      mask rem stack))
      (fail)))

(defun-nondeterministic verb-args6 (type voice mask0 phrase xlist-stack)
  (if (match? type '(_ be))
      (mvlet (((parsed-np mask rem stack) 
	       (verb-arg phrase xlist-stack)))
	     (returned-values (list 'arg (list 'dir 'pred parsed-np)) 
			      mask rem stack))
      (fail)))

(defun-nondeterministic verb-args7 (type voice mask0 phrase xlist-stack)
  (if (and (match? type '(_ have))
	   (eq 'active voice))
      (mvlet (((parsed-np mask rem stack) 
	       (verb-arg phrase xlist-stack)))
	     (returned-values (list 'arg (list 'dir 'pred parsed-np)) 
			      mask rem stack))
      (fail)))

(defun-nondeterministic verb-args8 (type voice mask0 phrase xlist-stack)
  (if (match? type (no-args))
      (returned-values (list 'arg '()) mask0 phrase xlist-stack)
      (fail)))


(defun-nondeterministic no-args ()
  '(_ intrans)) ;;( '(_ trans) '(_ ditrans) )


;;; maybe we can live w/o this.. try passing nil args back and see 
;;; if verb-mods take care of the pp ** check
(defun-nondeterministic verb-args9 (type voice mask0 phrase xlist-stack)
  (if (and (match? type '(aux be))
	   (eq voice 'passive))
      (returned-values (list 'arg '()) mask0 phrase xlist-stack)
    ;;  (either (mvlet (((parsed-pp mask rem stack)
;;		       (prep-p 'compl (s-all) phrase xlist-stack)))
;;		     (returned-values (list 'arg parsed-pp)
;;				      mask0 phrase xlist-stack))
	;;      (returned-values (list 'arg '()) mask0 phrase xlist-stack))
      (fail)))

(defun-nondeterministic object (mask phrase xlist-stack)
  ;; ** Check seems like mask should always be a vector!!
  (if (or (eq mask '_)
	  (equal mask '(_)))
      (setf mask (vector '_ '_ '_ '_)))
  (mvlet (((parsed-advs rem1 stack1)
	   (advs '() (minus (adv) mask) phrase xlist-stack)))
	 (mvlet (((case parsed-obj mask0 rem2 stack2)
		  (obj mask rem1 stack1)))
		(returned-values case (list parsed-advs parsed-obj)
				 rem2 stack2))))

;;
;; (defun-nondeterministic object (mask phrase xlist-stack)
;;   ;; ** Check seems like mask should always be a vector!!
;;   (if (or (eq mask '_)
;; 	  (equal mask '(_)))
;;       (let ((mask (vector '_ '_ '_ '_)))
;; 	(mvlet (((parsed-advs rem1 stack1)
;; 		 (advs '() (minus (adv) mask) phrase xlist-stack)))
;; 	       (mvlet (((case parsed-obj mask0 rem2 stack2)
;; 			(obj mask rem1 stack1)))
;; 		      (returned-values case (list parsed-advs parsed-obj)
;; 				       rem2 stack2))))
;;       	(mvlet (((parsed-advs rem1 stack1)
;; 		 (advs '() (minus (adv) mask) phrase xlist-stack)))
;; 	       (mvlet (((case parsed-obj mask0 rem2 stack2)
;; 			(obj mask rem1 stack1)))
;; 		      (returned-values case (list parsed-advs parsed-obj)
;; 				       rem2 stack2)))))


(defun-nondeterministic obj (mask0 phrase xlist-stack)
  (either (mvlet (((parsed-np mask rem stack) 
		   (verb-arg phrase xlist-stack)))
		 (returned-values 'ind (list 'dir parsed-np) mask rem stack))
	  (returned-values 'dir '() mask0 phrase xlist-stack))) 

(defun-nondeterministic pred-conj (phrase xlist-stack)
  (mvlet (((arg0 mask0 rem1 stack1)
	   (pred '_ phrase xlist-stack)))
	 (mvlet (((conj-val arg mask rem2 stack2)
		  (pred-rest arg0 mask0 rem1 stack1)))
		(returned-values conj-val arg mask rem2 stack2))))
		  
(defun-nondeterministic pred-rest (arg mask phrase xlist-stack)
  (either (mvlet (((conj0 parsed-conj arg1 rem1 stack1)
		   (conj arg '() phrase xlist-stack)))
		 (mvlet (((conj1 arg2 mask0 rem2 stack2)
			  (pred-conj rem1 stack1)))
			;;; *** check 
			(returned-values (list conj0 parsed-conj conj1) 
					 arg1 mask0 
					 rem2 stack2)))
	  (returned-values '() arg mask phrase xlist-stack)))

(defun-nondeterministic pred (pred-type phrase xlist-stack)
  (either (if (spec-eq? pred-type '_)
	      (mvlet (((adj-val mask rem stack)
		       (adj-phrase phrase xlist-stack)))
		     (if (not (null adj-val))
			 (returned-values adj-val mask rem stack)
			 (fail)))
	      (fail))
	  (if (spec-eq? pred-type 'neg)
	      (mvlet (((parsed-pp mask rem stack)
		       (prep-p 'compl (s-all) phrase xlist-stack)))
		     (returned-values parsed-pp mask rem stack))
	      (fail))
	  (if (spec-eq? pred-type '_)
	      (mvlet (((adv-val mask rem stack)
		       (adv-phrase (s-all) phrase xlist-stack)))
		     (returned-values adv-val mask rem stack))
	      (fail))))

(defun-nondeterministic advs (advs-so-far set phrase xlist-stack)
  (either (if (match? set (is-adv))
	      (mvlet (((adverb-val rem1 stack1)
		       (adverb phrase xlist-stack)))
		     (mvlet (((parsed-advs arg rem2 stack2)
			      (advs (append adverb-val advs-so-far)
				    set rem1 stack1)))
			    (returned-values parsed-advs '() rem2 stack2)))
	      (fail))
	  (returned-values advs-so-far '() phrase xlist-stack)))

(defun-nondeterministic adj-phrase (phrase xlist-stack)
  (either (mvlet (((adj-val rem stack)
		   (adjs phrase xlist-stack)))
		 (returned-values adj-val (empty) rem stack))
	  (mvlet (((pred-val mask rem stack)
		   (comp-phrase phrase xlist-stack)))
		 (returned-values pred-val mask rem stack))))

(defun-nondeterministic adv-phrase (set phrase xlist-stack)
  (either (mvlet (((non-term stack1) 
		   (virtual xlist-stack)))
		 (if (equal non-term (list 'adv-phrase set))
		     (returned-values (list 'adv_phrase non-term)
				      set phrase stack1)
		     (fail)))
	  (mvlet (((parsed-prep rem1 stack1)
		   (loc-pred phrase xlist-stack)))
		 (mvlet (((parsed-pp mask rem2 stack2)
			  (prep-p 'compl set  rem1 stack1)))
			(if (eq (prepostion parsed-pp) 'of)
			    (returned-values (list 'pp parsed-prep 
						   (noun-phrase parsed-pp))
					     mask rem2 stack2)
			    (fail))))))

(defun prepostion (prep-phrase)
  (second prep-phrase))

(defun noun-phrase (prep-phrase)
  (third prep-phrase))

				     

(defun-nondeterministic comp-phrase (phrase xlist-stack)
  (mvlet (((parsed-comp parsed-adj rem1 stack1)
	   (comp phrase xlist-stack)))
	 (mvlet (((parsed-arg agmt p-case type mask rem2 stack2)
		   (np 'compl (np-no-trace) rem1 stack1)))
		(if (match? (prep-case) p-case)
		    (returned-values (list 'comp parsed-comp parsed-adj) mask
				     rem2 stack2)
		    (fail)))))

(defun-nondeterministic comp (phrase xlist-stack)
  (either (mvlet (((parsed-comp rem1 stack1)
		   (comp-adv phrase xlist-stack)))
		 (mvlet (((parsed-adj rem2 stack2)
			  (adjs rem1 stack1)))
			(mvlet (((term rem3 stack3)
				 (terminal rem2 stack2)))
			       (if (eq term 'than)
				   (returned-values parsed-comp parsed-adj
						    rem3 stack3)
				   (fail)))))
	  (mvlet (((parsed-adj rem1 stack1)
		   (rel-adj phrase xlist-stack)))
		  (mvlet (((term rem2 stack2)
			   (terminal rem1 stack1)))
			 (if (eq term 'than)
			     (returned-values 'more parsed-adj
					      rem2 stack2)
			     (fail))))
	  (mvlet (((term rem1 stack1)
		   (terminal phrase xlist-stack)))
		 (if (eq term 'as)
		     (mvlet (((parsed-adj rem2 stack2)
			      (adjs rem1 stack1)))
			    (mvlet  (((term rem3 stack3)
				      (terminal rem2 stack2)))
				    (if (eq term 'as)
					(returned-values 'same parsed-adj
							 rem3 stack3)
					(fail))))
		     (fail)))))
  

(defun-nondeterministic verb-arg (phrase stack)
  (mvlet (((parsed-np agmt v-case type mask rem2 stack2) 
	   (np 'obj (s-all) phrase stack)))
	 (if (match? (verb-case) v-case)
	     (returned-values parsed-np mask rem2 stack2)
	     (fail))))



(defun-nondeterministic conj (left right phrase xlist-stack)
  (mvlet (((parsed-conj ctx0 ctx  rem stack)
	   (conj-val phrase xlist-stack)))
	 (returned-values `(conj (,parsed-conj ,ctx0))
			  `(conj (,parsed-conj ,ctx))
			  `(conj (,parsed-conj ,left ,right))
			  rem stack)))

