;;; -*- Syntax: common-lisp; Mode: LISP; Package: (parser :use common-lisp parser-util); Base: 10 -*- 

(in-package 'parser :use '(parser-util lisp))


;; push stack of xlist
;; xlist stack is made up 3-tuples namely: 
;; <context(gap/nogap), type(terminal/nonterminal), symbol(anything)>
(defun push-xlist (context type symbol stack)
  (append (list (list context type symbol))
	  stack))

;; pop stack of xlist
;; since the stack is made up of 3-tuples then it is easier and more elegant 
;; (I think) to have different functions to handle the stack

(defun top-context (stack)
  (if (null stack)
      (fail)
      (caar stack)))

(defun top-type (stack)
  (if (null stack)
      (fail)
      (cadar stack)))

(defun top-symbol (stack)
  (if (null stack)
      (fail)
      (caddar stack)))


(defun pop-stack (stack)
  (if (null stack)
      (fail)
      (cdr stack)))

;; vitrual
;; this function basically consumes one non-term of the xlist stack
;; based on two conditions
(defun virtual (stack)
  (if (not (null stack))
      (if (spec-eq? 'nonterminal (top-type stack))
	  (returned-values (top-symbol stack) (pop-stack stack))
	  (fail))
      (fail)))


;; Note that terminal is not used in this grammar
(defun-nondeterministic terminal (phrase stack)
  (either (terminal1 phrase stack)
	  (terminal2 phrase stack)))

;; consumes one terminal of the phrase based one some conditions
(defun-nondeterministic terminal1 (phrase stack)
  (if (gap stack) 
      (if (null phrase)
	  (fail)
	  (returned-values (car phrase) (cdr phrase) stack))
      (fail)))



;; consumes one element of the xlist stack
(defun terminal2 (phrase stack) 
  (if (not (null stack))
      (if (spec-eq? 'terminal (top-type stack))
	  (returned-values (top-symbol stack) phrase (pop-stack stack))
	  (fail))
      (fail)))

(defun-nondeterministic gap (stack)
  (either (gap1 stack)
	  (gap2 stack)))

(defun gap1 (stack)
  (if (not (null stack))
      (if (spec-eq? 'gap (top-context stack))
	  t
	  (fail))
      (fail)))

(defun gap2 (stack)
  (if (null stack)
      t
      (fail)))




;; Set definitions and operators for attachment control
;; these are all special predicates used for agreements

;; all vectors are 4-element vectors

;; normal form masks
(defun is-pp ()
  (vector 1 '_ '_ '_))

(defun is-pred ()
  (vector '_ 1 '_ '_))

(defun is-trace ()
  (vector '_ '_ 1 '_))

(defun is-adv ()
  (vector '_ '_ '_ 1))

(defun trace0 ()
  (returned-values (vector 0 0 1 0) 
		   (vector 0 0 0 0)))

(defun trace1 ()
  (vector '_ '_ 1 '_))

(defun adv ()
  (vector 0 0 0 1))

(defun empty ()
  (vector 0 0 0 0))

(defun np-all ()
  (vector 1 1 1 0))

(defun s-all ()
  (vector 1 0 1 1))

(defun np-no-trace ()
  (vector 1 1 0 0))

;; mask operations
(defun plus (vec1 vec2)
  (vector (or (vector-first vec1) (vector-first vec2))
	  (or (vector-second vec1) (vector-second vec2)) 
	  (or (vector-third vec1) (vector-third vec2))
	  (or (vector-fourth vec1) (vector-fourth vec2))))


(defun minus (vec1 vec2)
  (vector (anot (vector-first vec1) (vector-first vec2))
	  (anot (vector-second vec1) (vector-second vec2))
	  (anot (vector-third vec1) (vector-third vec2))
	  (anot (vector-fourth vec1) (vector-fourth vec2))))

(defun anot (x y)
  (and x (not y)))


;; Noun phrase position and features
(defun-nondeterministic role ()
  (either (returned-values 'subj '_ (vector 1 0 0))
	  (returned-values 'compl '_ (vector 0 '_ '_))
	  (returned-values 'obj '_ (vector 0 '_ '_))
	  (returned-values 'undef 'main (vector '_ 0 '_))
	  (returned-values 'undef 'aux (vector 0 '_ '_))
	  (returned-values 'undef 'decl (vector '_ '_ '_))
	  (returned-values 'nil '_ (vector '_ '_ '_))))


(defun subj-case ()
  (vector 1 0 0))

(defun verb-case ()
  (vector 0 1 0))

(defun prep-case ()
  (vector 0 0 1))

(defun compl-case () 
  (vector 0 '_ '_))


